"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonTemplateParser = void 0;
const types_1 = require("./types");
const errors_1 = require("./errors");
const constants_1 = require("./constants");
const lexer_1 = require("./lexer");
const utils_1 = require("./utils");
const engine_1 = require("./engine");
const EMPTY_EXPR = { type: types_1.SyntaxType.EMPTY };
class JsonTemplateParser {
    constructor(lexer, options) {
        this.lexer = lexer;
        this.options = options;
    }
    parse() {
        this.lexer.init();
        return this.parseStatementsExpr();
    }
    parseEndOfStatement(blockEnd) {
        if (this.lexer.matchEOT() || this.lexer.match(blockEnd)) {
            return;
        }
        if (this.lexer.match(';')) {
            this.lexer.ignoreTokens(1);
            return;
        }
        const currIdx = this.lexer.currentIndex();
        const nextTokenStart = this.lexer.lookahead().range[0];
        const code = this.lexer.getCode(currIdx, nextTokenStart);
        if (!code.includes('\n')) {
            this.lexer.throwUnexpectedToken();
        }
    }
    parseStatements(blockEnd) {
        let statements = [];
        while (!this.lexer.matchEOT() && !this.lexer.match(blockEnd)) {
            statements.push(this.parseStatementExpr());
            this.parseEndOfStatement(blockEnd);
        }
        return statements;
    }
    parseStatementsExpr(blockEnd) {
        return {
            type: types_1.SyntaxType.STATEMENTS_EXPR,
            statements: this.parseStatements(blockEnd),
        };
    }
    parseStatementExpr() {
        return this.parseBaseExpr();
    }
    parseAssignmentExpr() {
        const expr = this.parseNextExpr(types_1.OperatorType.ASSIGNMENT);
        if (expr.type === types_1.SyntaxType.PATH && this.lexer.match('=')) {
            this.lexer.ignoreTokens(1);
            const path = expr;
            if (!path.root || typeof path.root === 'object' || path.root === constants_1.DATA_PARAM_KEY) {
                throw new errors_1.JsonTemplateParserError('Invalid assignment path');
            }
            if (JsonTemplateParser.isRichPath(expr)) {
                throw new errors_1.JsonTemplateParserError('Invalid assignment path');
            }
            path.pathType = types_1.PathType.SIMPLE;
            return {
                type: types_1.SyntaxType.ASSIGNMENT_EXPR,
                value: this.parseBaseExpr(),
                path,
            };
        }
        return expr;
    }
    parseBaseExpr() {
        const startIdx = this.lexer.currentIndex();
        try {
            const expr = this.parseNextExpr(types_1.OperatorType.BASE);
            return expr;
        }
        catch (error) {
            const code = this.lexer.getCode(startIdx, this.lexer.currentIndex());
            if (error.message.includes('at')) {
                throw error;
            }
            throw new errors_1.JsonTemplateParserError(error.message + ' at ' + code);
        }
    }
    parseNextExpr(currentOperation) {
        switch (currentOperation) {
            case types_1.OperatorType.CONDITIONAL:
                return this.parseAssignmentExpr();
            case types_1.OperatorType.ASSIGNMENT:
                return this.parseCoalesceExpr();
            case types_1.OperatorType.COALESCING:
                return this.parseLogicalORExpr();
            case types_1.OperatorType.OR:
                return this.parseLogicalANDExpr();
            case types_1.OperatorType.AND:
                return this.parseEqualityExpr();
            case types_1.OperatorType.EQUALITY:
                return this.parseRelationalExpr();
            case types_1.OperatorType.RELATIONAL:
                return this.parseShiftExpr();
            case types_1.OperatorType.SHIFT:
                return this.parseAdditiveExpr();
            case types_1.OperatorType.ADDITION:
                return this.parseMultiplicativeExpr();
            case types_1.OperatorType.MULTIPLICATION:
                return this.parsePowerExpr();
            case types_1.OperatorType.POWER:
                return this.parseUnaryExpr();
            case types_1.OperatorType.UNARY:
                return this.parsePathAfterExpr();
            default:
                return this.parseConditionalExpr();
        }
    }
    parsePathPart() {
        if (this.lexer.match('.()')) {
            this.lexer.ignoreTokens(1);
        }
        else if (this.lexer.match('.') && this.lexer.match('(', 1)) {
            this.lexer.ignoreTokens(1);
            return this.parseBlockExpr();
        }
        else if (this.lexer.match('(')) {
            return this.parseFunctionCallExpr();
        }
        else if (this.lexer.matchPathPartSelector()) {
            return this.parseSelector();
        }
        else if (this.lexer.matchToArray()) {
            return this.parsePathOptions();
        }
        else if (this.lexer.match('[')) {
            return this.parseArrayFilterExpr();
        }
        else if (this.lexer.match('{')) {
            return this.parseObjectFiltersExpr();
        }
        else if (this.lexer.match('@') || this.lexer.match('#')) {
            return this.parsePathOptions();
        }
    }
    parsePathParts() {
        let parts = [];
        let newParts;
        while ((newParts = utils_1.CommonUtils.toArray(this.parsePathPart()))) {
            parts = parts.concat(newParts);
            if (newParts[0].type === types_1.SyntaxType.FUNCTION_CALL_EXPR) {
                break;
            }
        }
        return JsonTemplateParser.ignoreEmptySelectors(parts);
    }
    parseContextVariable() {
        this.lexer.ignoreTokens(1);
        if (!this.lexer.matchID()) {
            this.lexer.throwUnexpectedToken();
        }
        return this.lexer.value();
    }
    parsePathOptions() {
        const context = {};
        while (this.lexer.match('@') || this.lexer.match('#') || this.lexer.matchToArray()) {
            if (this.lexer.match('@')) {
                context.item = this.parseContextVariable();
                continue;
            }
            if (this.lexer.match('#')) {
                context.index = this.parseContextVariable();
                continue;
            }
            if (this.lexer.matchToArray()) {
                this.lexer.ignoreTokens(2);
                context.toArray = true;
            }
        }
        return {
            type: types_1.SyntaxType.PATH_OPTIONS,
            options: context,
        };
    }
    parsePathRoot(root) {
        if (root) {
            return root;
        }
        if (this.lexer.match('^')) {
            this.lexer.ignoreTokens(1);
            return constants_1.DATA_PARAM_KEY;
        }
        else if (this.lexer.matchID()) {
            return this.lexer.value();
        }
    }
    parsePathType() {
        if (this.lexer.matchSimplePath()) {
            this.lexer.ignoreTokens(1);
            return types_1.PathType.SIMPLE;
        }
        else if (this.lexer.matchRichPath()) {
            this.lexer.ignoreTokens(1);
            return types_1.PathType.RICH;
        }
        return this.options?.defaultPathType ?? types_1.PathType.RICH;
    }
    parsePath(options) {
        const pathType = this.parsePathType();
        let expr = {
            type: types_1.SyntaxType.PATH,
            root: this.parsePathRoot(options?.root),
            parts: this.parsePathParts(),
            pathType,
        };
        if (!expr.parts.length) {
            return expr;
        }
        return JsonTemplateParser.updatePathExpr(expr);
    }
    createArrayIndexFilterExpr(expr) {
        return {
            type: types_1.SyntaxType.ARRAY_INDEX_FILTER_EXPR,
            indexes: {
                type: types_1.SyntaxType.ARRAY_EXPR,
                elements: [expr],
            },
        };
    }
    createArrayFilterExpr(expr) {
        return {
            type: types_1.SyntaxType.ARRAY_FILTER_EXPR,
            filter: expr,
        };
    }
    parseSelector() {
        let selector = this.lexer.value();
        if (this.lexer.matchINT()) {
            return this.createArrayFilterExpr(this.createArrayIndexFilterExpr(this.parseLiteralExpr()));
        }
        let prop;
        if (this.lexer.match('*') || this.lexer.matchID() || this.lexer.matchTokenType(types_1.TokenType.STR)) {
            prop = this.lexer.lex();
        }
        return {
            type: types_1.SyntaxType.SELECTOR,
            selector: selector,
            prop,
        };
    }
    parseRangeFilterExpr() {
        if (this.lexer.match(':')) {
            this.lexer.ignoreTokens(1);
            return {
                type: types_1.SyntaxType.RANGE_FILTER_EXPR,
                toIdx: this.parseBaseExpr(),
            };
        }
        let fromExpr = this.parseBaseExpr();
        if (this.lexer.match(':')) {
            this.lexer.ignoreTokens(1);
            if (this.lexer.match(']')) {
                return {
                    type: types_1.SyntaxType.RANGE_FILTER_EXPR,
                    fromIdx: fromExpr,
                };
            }
            return {
                type: types_1.SyntaxType.RANGE_FILTER_EXPR,
                fromIdx: fromExpr,
                toIdx: this.parseBaseExpr(),
            };
        }
        return fromExpr;
    }
    parseArrayIndexFilterExpr(expr) {
        const parts = [];
        if (expr) {
            parts.push(expr);
            if (!this.lexer.match(']')) {
                this.lexer.expect(',');
            }
        }
        return {
            type: types_1.SyntaxType.ARRAY_INDEX_FILTER_EXPR,
            indexes: {
                type: types_1.SyntaxType.ARRAY_EXPR,
                elements: [
                    ...parts,
                    ...this.parseCommaSeparatedElements(']', () => this.parseSpreadExpr()),
                ],
            },
        };
    }
    parseArrayFilter() {
        if (this.lexer.matchSpread()) {
            return this.parseArrayIndexFilterExpr();
        }
        const expr = this.parseRangeFilterExpr();
        if (expr.type === types_1.SyntaxType.RANGE_FILTER_EXPR) {
            return expr;
        }
        return this.parseArrayIndexFilterExpr(expr);
    }
    parseObjectFilter() {
        let exclude = false;
        if (this.lexer.match('~')) {
            this.lexer.ignoreTokens(1);
            exclude = true;
        }
        // excluding is applicable only for index filters
        if (exclude || this.lexer.match('[')) {
            return {
                type: types_1.SyntaxType.OBJECT_INDEX_FILTER_EXPR,
                indexes: this.parseArrayExpr(),
                exclude,
            };
        }
        return {
            type: types_1.SyntaxType.OBJECT_FILTER_EXPR,
            filter: this.parseBaseExpr(),
        };
    }
    parseObjectFiltersExpr() {
        const objectFilters = [];
        const indexFilters = [];
        while (this.lexer.match('{')) {
            this.lexer.expect('{');
            const expr = this.parseObjectFilter();
            if (expr.type === types_1.SyntaxType.OBJECT_INDEX_FILTER_EXPR) {
                indexFilters.push(expr);
            }
            else {
                objectFilters.push(expr.filter);
            }
            this.lexer.expect('}');
            if (this.lexer.match('.') && this.lexer.match('{', 1)) {
                this.lexer.ignoreTokens(1);
            }
        }
        if (!objectFilters.length) {
            return indexFilters;
        }
        const objectFilter = {
            type: types_1.SyntaxType.OBJECT_FILTER_EXPR,
            filter: this.combineExpressionsAsBinaryExpr(objectFilters, types_1.SyntaxType.LOGICAL_AND_EXPR, '&&'),
        };
        return [objectFilter, ...indexFilters];
    }
    parseConditionalExpr() {
        const ifExpr = this.parseNextExpr(types_1.OperatorType.CONDITIONAL);
        if (this.lexer.match('?')) {
            this.lexer.ignoreTokens(1);
            const thenExpr = this.parseConditionalExpr();
            if (this.lexer.match(':')) {
                this.lexer.ignoreTokens(1);
                const elseExpr = this.parseConditionalExpr();
                return {
                    type: types_1.SyntaxType.CONDITIONAL_EXPR,
                    if: ifExpr,
                    then: thenExpr,
                    else: elseExpr,
                };
            }
            return {
                type: types_1.SyntaxType.CONDITIONAL_EXPR,
                if: ifExpr,
                then: thenExpr,
                else: {
                    type: types_1.SyntaxType.LITERAL,
                    tokenType: types_1.TokenType.UNDEFINED,
                },
            };
        }
        return ifExpr;
    }
    parseArrayFilterExpr() {
        this.lexer.expect('[');
        const filter = this.parseArrayFilter();
        this.lexer.expect(']');
        return {
            type: types_1.SyntaxType.ARRAY_FILTER_EXPR,
            filter,
        };
    }
    combineExpressionsAsBinaryExpr(values, type, op) {
        if (!values?.length) {
            throw new errors_1.JsonTemplateParserError('expected at least 1 expression');
        }
        if (values.length === 1) {
            return values[0];
        }
        return {
            type,
            op,
            args: [values.shift(), this.combineExpressionsAsBinaryExpr(values, type, op)],
        };
    }
    parseArrayCoalesceExpr() {
        this.lexer.ignoreTokens(1);
        const expr = this.parseArrayExpr();
        return this.combineExpressionsAsBinaryExpr(expr.elements, types_1.SyntaxType.LOGICAL_COALESCE_EXPR, '??');
    }
    parseCoalesceExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.COALESCING);
        if (this.lexer.match('??')) {
            return {
                type: types_1.SyntaxType.LOGICAL_COALESCE_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseCoalesceExpr()],
            };
        }
        return expr;
    }
    parseLogicalORExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.OR);
        if (this.lexer.match('||')) {
            return {
                type: types_1.SyntaxType.LOGICAL_OR_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseLogicalORExpr()],
            };
        }
        return expr;
    }
    parseLogicalANDExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.AND);
        if (this.lexer.match('&&')) {
            return {
                type: types_1.SyntaxType.LOGICAL_AND_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseLogicalANDExpr()],
            };
        }
        return expr;
    }
    parseEqualityExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.EQUALITY);
        if (this.lexer.match('==') ||
            this.lexer.match('!=') ||
            this.lexer.match('===') ||
            this.lexer.match('!==') ||
            this.lexer.match('^==') ||
            this.lexer.match('==^') ||
            this.lexer.match('^=') ||
            this.lexer.match('=^') ||
            this.lexer.match('$==') ||
            this.lexer.match('==$') ||
            this.lexer.match('$=') ||
            this.lexer.match('=$') ||
            this.lexer.match('*==') ||
            this.lexer.match('==*') ||
            this.lexer.match('*=') ||
            this.lexer.match('=*')) {
            return {
                type: types_1.SyntaxType.COMPARISON_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseEqualityExpr()],
            };
        }
        return expr;
    }
    parseRelationalExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.RELATIONAL);
        if (this.lexer.match('<') ||
            this.lexer.match('>') ||
            this.lexer.match('<=') ||
            this.lexer.match('>=') ||
            this.lexer.matchIN()) {
            return {
                type: this.lexer.matchIN() ? types_1.SyntaxType.IN_EXPR : types_1.SyntaxType.COMPARISON_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseRelationalExpr()],
            };
        }
        return expr;
    }
    parseShiftExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.SHIFT);
        if (this.lexer.match('>>') || this.lexer.match('<<')) {
            return {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseShiftExpr()],
            };
        }
        return expr;
    }
    parseAdditiveExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.ADDITION);
        if (this.lexer.match('+') || this.lexer.match('-')) {
            return {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseAdditiveExpr()],
            };
        }
        return expr;
    }
    parseMultiplicativeExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.MULTIPLICATION);
        if (this.lexer.match('*') || this.lexer.match('/') || this.lexer.match('%')) {
            return {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseMultiplicativeExpr()],
            };
        }
        return expr;
    }
    parsePowerExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.POWER);
        if (this.lexer.match('**')) {
            return {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parsePowerExpr()],
            };
        }
        return expr;
    }
    parseUnaryExpr() {
        if (this.lexer.match('!') ||
            this.lexer.match('-') ||
            this.lexer.matchTypeOf() ||
            this.lexer.matchAwait()) {
            return {
                type: types_1.SyntaxType.UNARY_EXPR,
                op: this.lexer.value(),
                arg: this.parseUnaryExpr(),
            };
        }
        return this.parseNextExpr(types_1.OperatorType.UNARY);
    }
    shouldSkipPathParsing(expr) {
        switch (expr.type) {
            case types_1.SyntaxType.DEFINITION_EXPR:
            case types_1.SyntaxType.ASSIGNMENT_EXPR:
            case types_1.SyntaxType.SPREAD_EXPR:
                return true;
            case types_1.SyntaxType.LITERAL:
            case types_1.SyntaxType.MATH_EXPR:
            case types_1.SyntaxType.COMPARISON_EXPR:
            case types_1.SyntaxType.ARRAY_EXPR:
            case types_1.SyntaxType.OBJECT_EXPR:
                if (this.lexer.match('(')) {
                    return true;
                }
                break;
            case types_1.SyntaxType.FUNCTION_EXPR:
                if (!this.lexer.match('(')) {
                    return true;
                }
                break;
        }
        return false;
    }
    parsePathAfterExpr() {
        let expr = this.parsePrimaryExpr();
        if (this.shouldSkipPathParsing(expr)) {
            return expr;
        }
        while (this.lexer.matchPathType() ||
            this.lexer.matchPathPartSelector() ||
            this.lexer.match('{') ||
            this.lexer.match('[') ||
            this.lexer.match('(')) {
            expr = this.parsePath({ root: expr });
        }
        return expr;
    }
    createLiteralExpr(token) {
        return {
            type: types_1.SyntaxType.LITERAL,
            value: token.value,
            tokenType: token.type,
        };
    }
    parseLiteralExpr() {
        return this.createLiteralExpr(this.lexer.lex());
    }
    parseIDPath() {
        const idParts = [];
        while (this.lexer.matchID()) {
            idParts.push(this.lexer.value());
            if (this.lexer.match('.') && this.lexer.matchID(1)) {
                this.lexer.ignoreTokens(1);
            }
        }
        if (!idParts.length) {
            this.lexer.throwUnexpectedToken();
        }
        return idParts.join('.');
    }
    parseObjectDefVars() {
        const vars = [];
        this.lexer.expect('{');
        while (!this.lexer.match('}')) {
            if (!this.lexer.matchID()) {
                throw new errors_1.JsonTemplateParserError('Invalid object vars');
            }
            vars.push(this.lexer.value());
            if (!this.lexer.match('}')) {
                this.lexer.expect(',');
            }
        }
        this.lexer.expect('}');
        if (vars.length === 0) {
            throw new errors_1.JsonTemplateParserError('Empty object vars');
        }
        return vars;
    }
    parseNormalDefVars() {
        const vars = [];
        if (!this.lexer.matchID()) {
            throw new errors_1.JsonTemplateParserError('Invalid normal vars');
        }
        vars.push(this.lexer.value());
        return vars;
    }
    parseDefinitionExpr() {
        const definition = this.lexer.value();
        let fromObject = this.lexer.match('{');
        let vars = fromObject ? this.parseObjectDefVars() : this.parseNormalDefVars();
        this.lexer.expect('=');
        return {
            type: types_1.SyntaxType.DEFINITION_EXPR,
            value: this.parseBaseExpr(),
            vars,
            definition,
            fromObject,
        };
    }
    parseFunctionCallArgs() {
        this.lexer.expect('(');
        const args = this.parseCommaSeparatedElements(')', () => this.parseSpreadExpr());
        this.lexer.expect(')');
        return args;
    }
    parseFunctionCallExpr() {
        let id;
        if (this.lexer.matchNew()) {
            this.lexer.ignoreTokens(1);
            id = 'new ' + this.parseIDPath();
        }
        return {
            type: types_1.SyntaxType.FUNCTION_CALL_EXPR,
            args: this.parseFunctionCallArgs(),
            id,
        };
    }
    parseFunctionDefinitionParam() {
        let spread = '';
        if (this.lexer.matchSpread()) {
            this.lexer.ignoreTokens(1);
            spread = '...';
            // rest param should be last param.
            if (!this.lexer.match(')', 1)) {
                this.lexer.throwUnexpectedToken();
            }
        }
        if (!this.lexer.matchID()) {
            this.lexer.throwUnexpectedToken();
        }
        return `${spread}${this.lexer.value()}`;
    }
    parseFunctionDefinitionParams() {
        this.lexer.expect('(');
        const params = this.parseCommaSeparatedElements(')', () => this.parseFunctionDefinitionParam());
        this.lexer.expect(')');
        return params;
    }
    parseFunctionExpr(asyncFn = false) {
        this.lexer.ignoreTokens(1);
        const params = this.parseFunctionDefinitionParams();
        this.lexer.expect('{');
        const statements = this.parseStatementsExpr('}');
        this.lexer.expect('}');
        return {
            type: types_1.SyntaxType.FUNCTION_EXPR,
            params,
            body: statements,
            async: asyncFn,
        };
    }
    parseObjectKeyExpr() {
        let key;
        if (this.lexer.match('[')) {
            this.lexer.ignoreTokens(1);
            key = this.parseBaseExpr();
            this.lexer.expect(']');
        }
        else if (this.lexer.matchID()) {
            key = this.lexer.value();
        }
        else if (this.lexer.matchTokenType(types_1.TokenType.STR)) {
            key = this.parseLiteralExpr();
        }
        else {
            this.lexer.throwUnexpectedToken();
        }
        return key;
    }
    parseObjectPropExpr() {
        let key;
        if (!this.lexer.matchSpread()) {
            key = this.parseObjectKeyExpr();
            this.lexer.expect(':');
        }
        const value = this.parseSpreadExpr();
        return {
            type: types_1.SyntaxType.OBJECT_PROP_EXPR,
            key,
            value,
        };
    }
    parseObjectExpr() {
        this.lexer.expect('{');
        const props = this.parseCommaSeparatedElements('}', () => this.parseObjectPropExpr());
        this.lexer.expect('}');
        return {
            type: types_1.SyntaxType.OBJECT_EXPR,
            props,
        };
    }
    parseCommaSeparatedElements(blockEnd, parseFn) {
        const elements = [];
        while (!this.lexer.match(blockEnd)) {
            elements.push(parseFn());
            if (!this.lexer.match(blockEnd)) {
                this.lexer.expect(',');
            }
        }
        return elements;
    }
    parseSpreadExpr() {
        if (this.lexer.matchSpread()) {
            this.lexer.ignoreTokens(1);
            return {
                type: types_1.SyntaxType.SPREAD_EXPR,
                value: this.parseBaseExpr(),
            };
        }
        return this.parseBaseExpr();
    }
    parseArrayExpr() {
        this.lexer.expect('[');
        const elements = this.parseCommaSeparatedElements(']', () => this.parseSpreadExpr());
        this.lexer.expect(']');
        return {
            type: types_1.SyntaxType.ARRAY_EXPR,
            elements,
        };
    }
    parseBlockExpr() {
        this.lexer.expect('(');
        let statements = this.parseStatements(')');
        this.lexer.expect(')');
        if (statements.length === 0) {
            throw new errors_1.JsonTemplateParserError('empty block is not allowed');
        }
        return {
            type: types_1.SyntaxType.BLOCK_EXPR,
            statements,
        };
    }
    parseAsyncFunctionExpr() {
        this.lexer.ignoreTokens(1);
        if (this.lexer.matchFunction()) {
            return this.parseFunctionExpr(true);
        }
        else if (this.lexer.matchLambda()) {
            return this.parseLambdaExpr(true);
        }
        this.lexer.throwUnexpectedToken();
    }
    parseLambdaExpr(asyncFn = false) {
        this.lexer.ignoreTokens(1);
        const expr = this.parseBaseExpr();
        return {
            type: types_1.SyntaxType.FUNCTION_EXPR,
            body: utils_1.CommonUtils.convertToStatementsExpr(expr),
            params: ['...args'],
            async: asyncFn,
        };
    }
    parseCompileTimeBaseExpr() {
        this.lexer.expect('{');
        this.lexer.expect('{');
        const expr = this.parseBaseExpr();
        this.lexer.expect('}');
        this.lexer.expect('}');
        return expr;
    }
    parseCompileTimeExpr() {
        this.lexer.expect('{');
        this.lexer.expect('{');
        const skipJsonify = this.lexer.matchCompileTimeExpr();
        const expr = skipJsonify ? this.parseCompileTimeBaseExpr() : this.parseBaseExpr();
        this.lexer.expect('}');
        this.lexer.expect('}');
        const exprVal = engine_1.JsonTemplateEngine.createAsSync(expr).evaluate({}, this.options?.compileTimeBindings);
        const template = skipJsonify ? exprVal : JSON.stringify(exprVal);
        return JsonTemplateParser.parseBaseExprFromTemplate(template);
    }
    parseNumber() {
        let val = this.lexer.value();
        if (this.lexer.match('.')) {
            val += this.lexer.value();
            if (this.lexer.matchINT()) {
                val += this.lexer.value();
            }
            return {
                type: types_1.SyntaxType.LITERAL,
                value: parseFloat(val),
                tokenType: types_1.TokenType.FLOAT,
            };
        }
        return {
            type: types_1.SyntaxType.LITERAL,
            value: parseInt(val, 10),
            tokenType: types_1.TokenType.INT,
        };
    }
    parseFloatingNumber() {
        const val = this.lexer.value() + this.lexer.value();
        return {
            type: types_1.SyntaxType.LITERAL,
            value: parseFloat(val),
            tokenType: types_1.TokenType.FLOAT,
        };
    }
    parseKeywordBasedExpr() {
        const token = this.lexer.lookahead();
        switch (token.value) {
            case types_1.Keyword.NEW:
                return this.parseFunctionCallExpr();
            case types_1.Keyword.LAMBDA:
                return this.parseLambdaExpr();
            case types_1.Keyword.ASYNC:
                return this.parseAsyncFunctionExpr();
            case types_1.Keyword.FUNCTION:
                return this.parseFunctionExpr();
            default:
                return this.parseDefinitionExpr();
        }
    }
    parsePrimaryExpr() {
        if (this.lexer.match(';')) {
            return EMPTY_EXPR;
        }
        if (this.lexer.matchTokenType(types_1.TokenType.LAMBDA_ARG)) {
            return {
                type: types_1.SyntaxType.LAMBDA_ARG,
                index: this.lexer.value(),
            };
        }
        if (this.lexer.matchKeyword()) {
            return this.parseKeywordBasedExpr();
        }
        if (this.lexer.matchINT()) {
            return this.parseNumber();
        }
        if (this.lexer.match('???')) {
            return this.parseArrayCoalesceExpr();
        }
        if (this.lexer.match('.') && this.lexer.matchINT(1) && !this.lexer.match('.', 2)) {
            return this.parseFloatingNumber();
        }
        if (this.lexer.matchLiteral()) {
            return this.parseLiteralExpr();
        }
        if (this.lexer.matchCompileTimeExpr()) {
            return this.parseCompileTimeExpr();
        }
        if (this.lexer.match('{')) {
            return this.parseObjectExpr();
        }
        if (this.lexer.match('[')) {
            return this.parseArrayExpr();
        }
        if (this.lexer.matchPath()) {
            return this.parsePath();
        }
        if (this.lexer.match('(')) {
            return this.parseBlockExpr();
        }
        return this.lexer.throwUnexpectedToken();
    }
    static pathContainsVariables(parts) {
        return parts
            .filter((part) => part.type === types_1.SyntaxType.PATH_OPTIONS)
            .some((part) => part.options?.index || part.options?.item);
    }
    static convertToBlockExpr(expr) {
        return {
            type: types_1.SyntaxType.FUNCTION_EXPR,
            block: true,
            body: utils_1.CommonUtils.convertToStatementsExpr(expr),
        };
    }
    static prependFunctionID(prefix, id) {
        return id ? prefix + '.' + id : prefix;
    }
    static ignoreEmptySelectors(parts) {
        return parts.filter((part) => !(part.type === types_1.SyntaxType.SELECTOR && part.selector === '.' && !part.prop));
    }
    static combinePathOptionParts(parts) {
        if (parts.length < 2) {
            return parts;
        }
        let newParts = [];
        for (let i = 0; i < parts.length; i++) {
            let currPart = parts[i];
            if (i < parts.length - 1 && parts[i + 1].type === types_1.SyntaxType.PATH_OPTIONS) {
                currPart.options = parts[i + 1].options;
                i++;
            }
            newParts.push(currPart);
        }
        return newParts;
    }
    static convertToFunctionCallExpr(fnExpr, pathExpr) {
        let lastPart = utils_1.CommonUtils.getLastElement(pathExpr.parts);
        if (lastPart?.type === types_1.SyntaxType.SELECTOR) {
            const selectorExpr = lastPart;
            if (selectorExpr.selector === '.' && selectorExpr.prop?.type === types_1.TokenType.ID) {
                pathExpr.parts.pop();
                fnExpr.id = selectorExpr.prop.value;
                fnExpr.dot = true;
            }
        }
        if (!pathExpr.parts.length && pathExpr.root && typeof pathExpr.root !== 'object') {
            fnExpr.id = this.prependFunctionID(pathExpr.root, fnExpr.id);
            fnExpr.dot = false;
        }
        else {
            fnExpr.object = pathExpr;
        }
        return fnExpr;
    }
    static isArrayFilterExpressionSimple(expr) {
        if (expr.filter.type !== types_1.SyntaxType.ARRAY_INDEX_FILTER_EXPR) {
            return false;
        }
        const filter = expr.filter;
        if (filter.indexes.elements.length > 1) {
            return false;
        }
        return true;
    }
    static isSimplePathPart(part) {
        if (part.type === types_1.SyntaxType.SELECTOR) {
            const expr = part;
            return expr.selector === '.' && !!expr.prop && expr.prop.type !== types_1.TokenType.PUNCT;
        }
        if (part.type === types_1.SyntaxType.ARRAY_FILTER_EXPR) {
            return this.isArrayFilterExpressionSimple(part);
        }
        return false;
    }
    static isSimplePath(pathExpr) {
        return pathExpr.parts.every((part) => this.isSimplePathPart(part));
    }
    static isRichPath(pathExpr) {
        if (!pathExpr.parts.length) {
            return false;
        }
        return !this.isSimplePath(pathExpr);
    }
    static updatePathExpr(pathExpr) {
        if (pathExpr.parts.length > 1 && pathExpr.parts[0].type === types_1.SyntaxType.PATH_OPTIONS) {
            pathExpr.options = pathExpr.parts[0].options;
            pathExpr.parts.shift();
        }
        const shouldConvertAsBlock = JsonTemplateParser.pathContainsVariables(pathExpr.parts);
        let lastPart = utils_1.CommonUtils.getLastElement(pathExpr.parts);
        let fnExpr;
        if (lastPart?.type === types_1.SyntaxType.FUNCTION_CALL_EXPR) {
            fnExpr = pathExpr.parts.pop();
        }
        lastPart = utils_1.CommonUtils.getLastElement(pathExpr.parts);
        if (lastPart?.type === types_1.SyntaxType.PATH_OPTIONS) {
            pathExpr.parts.pop();
            pathExpr.returnAsArray = lastPart.options?.toArray;
        }
        pathExpr.parts = JsonTemplateParser.combinePathOptionParts(pathExpr.parts);
        let expr = pathExpr;
        if (fnExpr) {
            expr = JsonTemplateParser.convertToFunctionCallExpr(fnExpr, pathExpr);
        }
        if (shouldConvertAsBlock) {
            expr = JsonTemplateParser.convertToBlockExpr(expr);
            pathExpr.pathType = types_1.PathType.RICH;
        }
        else if (this.isRichPath(pathExpr)) {
            pathExpr.pathType = types_1.PathType.RICH;
        }
        return expr;
    }
    static parseBaseExprFromTemplate(template) {
        const lexer = new lexer_1.JsonTemplateLexer(template);
        const parser = new JsonTemplateParser(lexer);
        return parser.parseBaseExpr();
    }
}
exports.JsonTemplateParser = JsonTemplateParser;
